package cn.com.qmth.am.service.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import com.qmth.boot.core.exception.StatusException;
import com.qmth.boot.tools.models.ByteArray;

import cn.com.qmth.am.bean.ImagePosition;
import cn.com.qmth.am.bean.ImageSize;
import cn.com.qmth.am.bean.OcrDto;
import cn.com.qmth.am.bean.StudentScoreVo;
import cn.com.qmth.am.bean.StudentVo;
import cn.com.qmth.am.bean.TrackPosition;
import cn.com.qmth.am.config.SysProperty;
import cn.com.qmth.am.entity.QuestionEntity;
import cn.com.qmth.am.enums.AnswerRangeType;
import cn.com.qmth.am.enums.DataStatus;
import cn.com.qmth.am.service.QuestionService;
import cn.com.qmth.am.service.StmmsService;
import cn.com.qmth.am.service.StudentScoreService;
import cn.com.qmth.am.service.StudentService;
import cn.com.qmth.am.utils.ImageUtil;

@Service
public class StudentServiceImpl implements StudentService {

    private static final Logger log = LoggerFactory.getLogger(StudentService.class);

    @Autowired
    private StudentService studentService;

    @Autowired
    private SysProperty sysProperty;

    @Autowired
    private StmmsService stmmsService;

    @Autowired
    private StudentScoreService studentScoreService;

    @Autowired
    private QuestionService questionService;

    @Override
    public void importStudent() {
        List<QuestionEntity> qs = questionService.findToDispose();
        if (CollectionUtils.isEmpty(qs)) {
            return;
        }
        Map<String, List<QuestionEntity>> qmap = new LinkedHashMap<>();
        for (QuestionEntity q : qs) {
            String key = q.getExamId() + "-" + q.getSubjectCode();
            List<QuestionEntity> tem = qmap.get(key);
            if (tem == null) {
                tem = new ArrayList<>();
                qmap.put(key, tem);
            }
            tem.add(q);
        }
        for (List<QuestionEntity> qlist : qmap.values()) {
            List<StudentVo> students = stmmsService.getUploadStudent(qlist.get(0).getExamId(),
                    qlist.get(0).getSubjectCode(), sysProperty.getStudentCount());
            if (CollectionUtils.isEmpty(students)) {
                continue;
            }
            updateQuestionsStatus(qlist, DataStatus.PROCESSING);
            try {
                studentScoreService.saveByQuestion(qlist, students);
                updateQuestionsStatus(qlist, DataStatus.SUCCESS);
            } catch (Exception e) {
                log.error("同步考生出错", e);
                updateQuestionsStatus(qlist, DataStatus.FAILED);
            }
        }
    }

    private void updateQuestionsStatus(List<QuestionEntity> qlist, DataStatus st) {
        for (QuestionEntity q : qlist) {
            questionService.updateStatus(q.getId(), st);
        }
    }

    @Override
    public void buildImage(OcrDto ocrDto) {
        if (DataStatus.WAITING.equals(ocrDto.getScore().getAnswerStatus())
                || DataStatus.FAILED.equals(ocrDto.getScore().getAnswerStatus())) {
            if (AnswerRangeType.TRACK.equals(ocrDto.getQuetion().getAnswerRangeType())
                    && CollectionUtils.isEmpty(ocrDto.getScore().getImagePosition())) {
                return;
            } else if (AnswerRangeType.ALL.equals(ocrDto.getQuetion().getAnswerRangeType())
                    && ocrDto.getScore().getSheetCount() == null) {
                return;
            }
            studentService.createSlice(ocrDto);
        }
    }

    @Override
    public void createSlice(OcrDto ocrDto) {
        if (ocrDto.getQuetion() == null) {
            studentScoreService.updateAnswerErr(ocrDto.getScore().getId(), "未找到试题信息");
            return;
        }
        try {
            studentScoreService.createSlice(ocrDto);
        } catch (Exception e) {
            if (e instanceof StatusException) {
                studentScoreService.updateAnswerErr(ocrDto.getScore().getId(), e.getMessage());
            } else {
                log.error("系统异常", e);
                studentScoreService.updateAnswerErr(ocrDto.getScore().getId(), "系统异常");
            }
        }
    }

    @Transactional
    @Override
    public void reset(Long examId, String subjectCode) {
        studentScoreService.removeBy(examId, subjectCode);
    }

    @Transactional
    @Override
    public void clear(Long examId, String subjectCode) {
        studentScoreService.removeBy(examId, subjectCode);
        questionService.removeBy(examId, subjectCode);
    }

    @Override
    public void updateMarkingScore(QuestionEntity q, StudentScoreVo score) {
        List<Double> list = stmmsService.getMarkScore(score.getStudentId(), q.getMainNumber(), q.getSubNumber());
        if (CollectionUtils.isEmpty(list)) {
            return;
        }
        score.setMarkingScore(list.get(0));
        if (AnswerRangeType.TRACK.equals(q.getAnswerRangeType())) {
            score.setIps(getImagePosition(q.getExamId(), score.getExamNumber(), score.getStudentId(),
                    q.getMainNumber() + "." + q.getSubNumber()));
            studentScoreService.updateMarkingScoreAndTrack(score);
        } else {
            studentScoreService.updateMarkingScore(score);
        }
    }

    private List<ImagePosition> getImagePosition(Long examId, String examNumber, Long studentId,
            String questionNumber) {
        List<ImagePosition> ret = new ArrayList<>();
        List<TrackPosition> tps = stmmsService.getTrackPosition(studentId, questionNumber);
        Map<Integer, Integer> answerImages = new HashMap<>();
        Set<String> set = new HashSet<>();
        for (TrackPosition tp : tps) {
            Integer width = getWidth(examId, examNumber, tp.getOffsetIndex(), answerImages);
            Boolean left = tp.getOffsetX() < (width / 2);
            String key = left.toString() + tp.getOffsetIndex();
            if (!set.contains(key)) {
                ImagePosition ip = new ImagePosition();
                ip.setLeft(left);
                ip.setPageIndex(tp.getOffsetIndex());
                ret.add(ip);
            }
        }
        return ret;
    }

    private Integer getWidth(Long examId, String examNumber, Integer pageIndex, Map<Integer, Integer> answerImages) {
        Integer ret = answerImages.get(pageIndex);
        if (ret != null) {
            return ret;
        }
        String url = getImageUrl(examId, examNumber, pageIndex);
        url = url.replace("https", "http");
        try {
            ImageSize is = ImageUtil.getSize(ByteArray.fromUrl(url).value());
            answerImages.put(pageIndex, is.getWidth());
            ret = is.getWidth();
            return ret;
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private String getImageUrl(Long examId, String examNumber, Integer pageIndex) {
        return sysProperty.getImageServer() + "/"
                + getMarkingCloudPath(examId, getSuffix(examNumber), examNumber, pageIndex, "jpg");
    }

    private static String getSuffix(String input) {
        return StringUtils.trimToEmpty(input).substring(Math.max(0, input.length() - 3));
    }

    private static String getMarkingCloudPath(Object... param) {
        return String.format("sheet/%d/%s/%s-%d.%s", param);
    }

}